@@ -26,8 +26,10 @@ extern void struct_passwd_freep(struct passwd **var);
 extern struct spwd *struct_shadow_free(struct spwd *var);
 extern void struct_shadow_freep(struct spwd **var);
 
+/* All returning structs and strings, if not "const", need to be free'd
+   after usage. This are especially name, pw, sp and error. */
 extern int pwaccess_check_expired(const char *user, long *daysleft,
-				  bool *pwchangeable, char ** error);
+				  bool *pwchangeable, char **error);
 extern int pwaccess_get_account_name(int64_t uid, char **name, char **error);
 extern int pwaccess_get_user_record(int64_t uid, const char *user,
 				    struct passwd **pw, struct spwd **sp,
