@@ -6,11 +6,12 @@
 
 #include "check_caller_perms.h"
 
-/* Don't allow access if query does not come from root
-   and the result is not the one of the calling user.
-   Exception: if the peer uid is in the list of exceptions.
-   "Lex mariadb": user mysql/mariadb needs to authenticate other
-   users to allow access to the database. */
+/*  Do not allow access if the query does not originate from root
+    or the entry does not belong to the calling user.
+    Exception: if the peer uid is in the list of exceptions.
+    "Lex mariadb": user mysql/mariadb needs to authenticate as
+    database user so that the database user can get access to the
+    database. */
 bool
 check_caller_perms(uid_t peer_uid, uid_t target_uid, uid_t *allowed)
 {
