@@ -2,7 +2,6 @@
 
 #include "config.h"
 
-#include <assert.h>
 #include <limits.h>
 #include <getopt.h>
 #include <stdlib.h>
@@ -19,15 +18,16 @@
 #include "basics.h"
 #include "mkdir_p.h"
 #include "verify.h"
+#include "check_caller_perms.h"
 #include "varlink-service-common.h"
+#include "read_config.h"
+#include "context.h"
 
 #include "varlink-org.openSUSE.pwaccess.h"
 
 #define USEC_PER_SEC  ((uint64_t) 1000000ULL)
 #define DEFAULT_EXIT_USEC (30*USEC_PER_SEC)
 
-static int socket_activation = false;
-
 static int
 error_user_not_found(sd_varlink *link, int64_t uid, const char *name, int errcode)
 {
@@ -131,8 +131,9 @@ vl_method_get_account_name(sd_varlink *link, sd_json_variant *parameters,
 static int
 vl_method_get_user_record(sd_varlink *link, sd_json_variant *parameters,
 			  sd_varlink_method_flags_t _unused_(flags),
-			  void _unused_(*userdata))
+			  void *userdata)
 {
+  struct context_t *ctx = userdata;
   _cleanup_(sd_json_variant_unrefp) sd_json_variant *passwd = NULL;
   _cleanup_(sd_json_variant_unrefp) sd_json_variant *shadow = NULL;
   _cleanup_(sd_json_variant_unrefp) sd_json_variant *result = NULL;
@@ -206,10 +207,10 @@ vl_method_get_user_record(sd_varlink *link, sd_json_variant *parameters,
 
   /* Don't return password if query does not come from root
      and result is not the one of the calling user */
-  if (peer_uid != 0 && pw->pw_uid != peer_uid)
+  if (!check_caller_perms(peer_uid, pw->pw_uid, ctx->cfg.allow_get_user_record))
     {
-      log_msg(LOG_DEBUG, "GetUserRecord: peer UID not 0 and UID not equal to peer UID (%i,%i)",
-	      peer_uid, pw->pw_uid);
+      log_msg(LOG_DEBUG, "Peer UID %u is not allowed to access data of '%s'",
+	  peer_uid, pw->pw_name);
       pw->pw_passwd = NULL;
       complete = false;
       /* no shadow entries for others */
@@ -303,8 +304,9 @@ vl_method_get_user_record(sd_varlink *link, sd_json_variant *parameters,
 static int
 vl_method_verify_password(sd_varlink *link, sd_json_variant *parameters,
 			  sd_varlink_method_flags_t _unused_(flags),
-			  void _unused_(*userdata))
+			  void *userdata)
 {
+  struct context_t *ctx = userdata;
   _cleanup_(parameters_free) struct parameters p = {
     .name = NULL,
     .password = NULL,
@@ -349,16 +351,14 @@ vl_method_verify_password(sd_varlink *link, sd_json_variant *parameters,
   if (pw == NULL)
     return error_user_not_found(link, -1, p.name, errno);
 
-  /* Don't verify password if query does not come from root
-     and result is not the one of the calling user */
-  if (peer_uid != 0 && pw->pw_uid != peer_uid)
+  if (!check_caller_perms(peer_uid, pw->pw_uid, ctx->cfg.allow_verify_password))
     {
       _cleanup_free_ char *error = NULL;
 
-      if (asprintf(&error, "Peer UID (%i) not 0 and peer UID not equal to UID",
-		   peer_uid) < 0)
+      if (asprintf(&error, "Peer UID %u is not allowed to verify password of '%s'",
+		   peer_uid, p.name) < 0)
 	error = NULL;
-      log_msg(LOG_ERR, "VerifyPassword: %s", stroom(error));
+      log_msg(LOG_ERR, "%s", stroom(error));
       return sd_varlink_errorbo(link, "org.openSUSE.pwaccess.InternalError",
 				SD_JSON_BUILD_PAIR_BOOLEAN("Success", false),
 				SD_JSON_BUILD_PAIR_STRING("ErrorMsg", stroom(error)));
@@ -431,8 +431,9 @@ vl_method_verify_password(sd_varlink *link, sd_json_variant *parameters,
 static int
 vl_method_expired_check(sd_varlink *link, sd_json_variant *parameters,
 			sd_varlink_method_flags_t _unused_(flags),
-			void _unused_(*userdata))
+			void *userdata)
 {
+  struct context_t *ctx = userdata;
   _cleanup_(parameters_free) struct parameters p = {
     .name = NULL,
     .password = NULL,
@@ -478,12 +479,12 @@ vl_method_expired_check(sd_varlink *link, sd_json_variant *parameters,
 
   /* Don't verify password if query does not come from root
      and result is not the one of the calling user */
-  if (peer_uid != 0 && pw->pw_uid != peer_uid)
+  if (!check_caller_perms(peer_uid, pw->pw_uid, ctx->cfg.allow_expired_check))
     {
       _cleanup_free_ char *error = NULL;
 
-      if (asprintf(&error, "Peer UID (%i) not 0 and peer UID not equal to UID",
-		   peer_uid) < 0)
+      if (asprintf(&error, "Peer UID %u is not allowed to access data of '%s'",
+		   peer_uid, p.name) < 0)
 	error = NULL;
       log_msg(LOG_ERR, "ExpiredCheck: %s", stroom(error));
       return sd_varlink_errorbo(link, "org.openSUSE.pwaccess.InternalError",
@@ -586,7 +587,7 @@ varlink_event_loop_with_idle(sd_event *e, sd_varlink_server *s)
 }
 
 static int
-run_varlink (void)
+run_varlink(bool socket_activation, struct context_t *ctx)
 {
   int r;
   _cleanup_(sd_event_unrefp) sd_event *event = NULL;
@@ -600,27 +601,28 @@ run_varlink (void)
       return r;
     }
 
-  r = sd_event_new (&event);
+  r = sd_event_new(&event);
   if (r < 0)
     {
-      log_msg (LOG_ERR, "Failed to create new event: %s",
-	       strerror (-r));
+      log_msg(LOG_ERR, "Failed to create new event: %s",
+	      strerror(-r));
+      return r;
+    }
+  ctx->loop = TAKE_PTR(event);
+
+  r = sd_varlink_server_new(&varlink_server, SD_VARLINK_SERVER_ACCOUNT_UID|SD_VARLINK_SERVER_INHERIT_USERDATA|SD_VARLINK_SERVER_INPUT_SENSITIVE);
+  if (r < 0)
+    {
+      log_msg(LOG_ERR, "Failed to allocate varlink server: %s",
+	      strerror (-r));
       return r;
     }
 
-  r = sd_varlink_server_new (&varlink_server, SD_VARLINK_SERVER_ACCOUNT_UID|SD_VARLINK_SERVER_INHERIT_USERDATA|SD_VARLINK_SERVER_INPUT_SENSITIVE);
+  r = sd_varlink_server_set_description(varlink_server, "pwaccessd");
   if (r < 0)
     {
-      log_msg (LOG_ERR, "Failed to allocate varlink server: %s",
-	       strerror (-r));
-      return r;
-    }
-
-  r = sd_varlink_server_set_description (varlink_server, "pwaccessd");
-  if (r < 0)
-    {
-      log_msg (LOG_ERR, "Failed to set varlink server description: %s",
-	       strerror (-r));
+      log_msg(LOG_ERR, "Failed to set varlink server description: %s",
+	      strerror(-r));
       return r;
     }
 
@@ -636,15 +638,17 @@ run_varlink (void)
       return r;
     }
 
-  r = sd_varlink_server_bind_method_many (varlink_server,
-					  "org.openSUSE.pwaccess.GetAccountName", vl_method_get_account_name,
-					  "org.openSUSE.pwaccess.GetUserRecord",  vl_method_get_user_record,
-					  "org.openSUSE.pwaccess.VerifyPassword", vl_method_verify_password,
-					  "org.openSUSE.pwaccess.ExpiredCheck",   vl_method_expired_check,
-					  "org.openSUSE.pwaccess.GetEnvironment", vl_method_get_environment,
-					  "org.openSUSE.pwaccess.Ping",           vl_method_ping,
-					  "org.openSUSE.pwaccess.Quit",           vl_method_quit,
-					  "org.openSUSE.pwaccess.SetLogLevel",    vl_method_set_log_level);
+  sd_varlink_server_set_userdata(varlink_server, ctx);
+
+  r = sd_varlink_server_bind_method_many(varlink_server,
+					 "org.openSUSE.pwaccess.GetAccountName", vl_method_get_account_name,
+					 "org.openSUSE.pwaccess.GetUserRecord",  vl_method_get_user_record,
+					 "org.openSUSE.pwaccess.VerifyPassword", vl_method_verify_password,
+					 "org.openSUSE.pwaccess.ExpiredCheck",   vl_method_expired_check,
+					 "org.openSUSE.pwaccess.GetEnvironment", vl_method_get_environment,
+					 "org.openSUSE.pwaccess.Ping",           vl_method_ping,
+					 "org.openSUSE.pwaccess.Quit",           vl_method_quit,
+					 "org.openSUSE.pwaccess.SetLogLevel",    vl_method_set_log_level);
   if (r < 0)
     {
       log_msg(LOG_ERR, "Failed to bind Varlink methods: %s",
@@ -652,19 +656,17 @@ run_varlink (void)
       return r;
     }
 
-  sd_varlink_server_set_userdata (varlink_server, event);
-
-  r = sd_varlink_server_attach_event (varlink_server, event, SD_EVENT_PRIORITY_NORMAL);
+  r = sd_varlink_server_attach_event(varlink_server, ctx->loop, SD_EVENT_PRIORITY_NORMAL);
   if (r < 0)
     {
-      log_msg (LOG_ERR, "Failed to attach to event: %s", strerror (-r));
+      log_msg(LOG_ERR, "Failed to attach to event: %s", strerror(-r));
       return r;
     }
 
-  r = sd_varlink_server_listen_auto (varlink_server);
+  r = sd_varlink_server_listen_auto(varlink_server);
   if (r < 0)
     {
-      log_msg (LOG_ERR, "Failed to listen: %s", strerror (-r));
+      log_msg (LOG_ERR, "Failed to listen: %s", strerror(-r));
       return r;
     }
 
@@ -674,23 +676,23 @@ run_varlink (void)
       r = sd_varlink_server_listen_address(varlink_server, _VARLINK_PWACCESS_SOCKET, 0666);
       if (r < 0)
 	{
-	  log_msg (LOG_ERR, "Failed to bind to Varlink socket: %s", strerror (-r));
+	  log_msg(LOG_ERR, "Failed to bind to Varlink socket: %s", strerror(-r));
 	  return r;
 	}
     }
 
   announce_ready();
   if (socket_activation)
-    r = varlink_event_loop_with_idle(event, varlink_server);
+    r = varlink_event_loop_with_idle(ctx->loop, varlink_server);
   else
-    r = sd_event_loop(event);
+    r = sd_event_loop(ctx->loop);
   announce_stopping();
 
   return r;
 }
 
 static void
-print_help (void)
+print_help(void)
 {
   printf("pwaccessd - manage passwd and shadow\n");
 
@@ -701,9 +703,25 @@ print_help (void)
   printf("      --version  Print program version\n");
 }
 
-int
-main (int argc, char **argv)
+static void
+struct_context_free(struct context_t *var)
 {
+  struct_config_free(&(var->cfg));
+  sd_event_unrefp(&(var->loop));
+}
+
+
+int
+main(int argc, char **argv)
+{
+  int socket_activation = false;
+  _cleanup_(struct_context_free) struct context_t ctx = { {NULL, NULL, NULL}, NULL };
+  econf_err error = read_config(&ctx.cfg);
+
+  if (error != ECONF_SUCCESS)
+    log_msg(LOG_NOTICE, "Error reading config file: %s",
+	    econf_errString(error));
+
   while (1)
     {
       int c;
@@ -758,7 +776,7 @@ main (int argc, char **argv)
 
   log_msg (LOG_INFO, "Starting pwaccessd (%s) %s...", PACKAGE, VERSION);
 
-  int r = run_varlink ();
+  int r = run_varlink (socket_activation, &ctx);
   if (r < 0)
     {
       log_msg (LOG_ERR, "ERROR: varlink loop failed: %s", strerror (-r));
