@@ -26,6 +26,7 @@
 #include "files.h"
 #include "verify.h"
 #include "chfn_checks.h"
+#include "check_caller_perms.h"
 
 #include "varlink-org.openSUSE.pwupd.h"
 
@@ -96,10 +97,6 @@ struct parameters {
   int flags;
   sd_json_variant *content_passwd;
   sd_json_variant *content_shadow;
-  /* run_as_user:
-     0: let PAM module decide as what we run
-     1: PAM modules should assume we run as user even if geteuid() returns 0 */
-  int run_as_user;
   sd_varlink *link;
 };
 
@@ -123,6 +120,9 @@ static sd_json_variant *send_v = NULL;
 static pthread_mutex_t mut = PTHREAD_MUTEX_INITIALIZER;
 static pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
 static char *answer = NULL;
+static bool got_answer = false;
+static bool broadcast_called = false;
+
 
 static int
 varlink_conv(int num_msg, const struct pam_message **msgm,
@@ -147,30 +147,31 @@ varlink_conv(int num_msg, const struct pam_message **msgm,
         case PAM_PROMPT_ECHO_ON:
         case PAM_PROMPT_ECHO_OFF:
 	  pthread_mutex_lock(&mut);
-	  if (send_v != NULL)
-	    sd_json_variant_unref(send_v);
-	  send_v = NULL;
+	  send_v = sd_json_variant_unref(send_v);
 	  r = sd_json_variant_merge_objectbo(&send_v,
 					     SD_JSON_BUILD_PAIR_INTEGER("msg_style", msgm[count]->msg_style),
 					     SD_JSON_BUILD_PAIR("message", SD_JSON_BUILD_STRING(msgm[count]->msg)));
 	  if (r < 0)
 	    log_msg(LOG_ERR, "Failed to build send_v list: %s\n", strerror(-r));
+	  broadcast_called = true;
 	  pthread_cond_broadcast(&cond);
 	  pthread_mutex_unlock(&mut);
 
-	  /* waiting for answer */
-	  pthread_mutex_lock(&mut);
 	  *response = calloc(num_msg, sizeof(struct pam_response));
 	  if (*response == NULL)
 	    {
 	      log_msg(LOG_ERR, "Out of memory!");
-	      pthread_mutex_unlock(&mut);
 	      return PAM_BUF_ERR;
 	    }
-	  log_msg(LOG_DEBUG, "varlink_conv: calling pthread_cond_wait");
-	  pthread_cond_wait(&cond, &mut);
+
+	  /* waiting for answer */
+	  pthread_mutex_lock(&mut);
+	  log_msg(LOG_DEBUG, "varlink_conv: wait for answer from client");
+	  while (!got_answer)
+	    pthread_cond_wait(&cond, &mut);
 	  response[0]->resp_retcode = 0;
-	  response[0]->resp = answer;
+	  response[0]->resp = TAKE_PTR(answer);
+	  got_answer = false;
 	  pthread_mutex_unlock(&mut);
 	  log_msg(LOG_DEBUG, "varlink_conv: after mutex");
           break;
@@ -199,9 +200,9 @@ static void *
 broadcast_and_return(intptr_t r)
 {
   pthread_mutex_lock(&mut);
+  broadcast_called = true;
   pthread_cond_broadcast(&cond);
   pthread_mutex_unlock(&mut);
-
   return (void *)r;
 }
 
@@ -224,7 +225,6 @@ run_pam_auth(void *arg)
     .content_passwd = NULL,
     .content_shadow = NULL,
     .link = param->link,
-    .run_as_user = param->run_as_user,
   };
   const struct pam_conv conv = {
     varlink_conv,
@@ -241,16 +241,6 @@ run_pam_auth(void *arg)
       return broadcast_and_return(r);
     }
 
-  if (p.run_as_user)
-    {
-      r = pam_putenv(pamh, "PAM_NO_ROOT=1");
-      if (r != PAM_SUCCESS)
-	{
-	  log_msg(LOG_ERR, "pam_putenv(\"PAM_NO_ROOT=1\") failed: %s", pam_strerror(NULL, r));
-	  return broadcast_and_return(r);
-	}
-    }
-
   r = pam_authenticate(pamh, 0);
   if (r != PAM_SUCCESS)
     {
@@ -369,6 +359,18 @@ run_pam_auth(void *arg)
 }
 
 static pthread_t pam_thread;
+static bool pam_thread_is_valid = false;
+
+static int
+thread_is_running(void)
+{
+  if (pam_thread_is_valid)
+    {
+      int r = pthread_kill(pam_thread, 0);
+      return -r;
+    }
+  return -ENOENT;
+}
 
 static int
 vl_method_chfn(sd_varlink *link, sd_json_variant *parameters,
@@ -391,7 +393,6 @@ vl_method_chfn(sd_varlink *link, sd_json_variant *parameters,
     .content_passwd = NULL,
     .content_shadow = NULL,
     .link = link,
-    .run_as_user = 0,
   };
   static const sd_json_dispatch_field dispatch_table[] = {
     { "userName",  SD_JSON_VARIANT_STRING,  sd_json_dispatch_string, offsetof(struct parameters, name),       SD_JSON_MANDATORY},
@@ -409,13 +410,28 @@ vl_method_chfn(sd_varlink *link, sd_json_variant *parameters,
 
   log_msg(LOG_INFO, "Varlink method \"chfn\" called...");
 
-  r = sd_varlink_get_peer_uid(link, &peer_uid);
-  if (r < 0)
+  /* If there is already a thread running, quit with error.
+     The conv method needs to be called to continue. */
+  r = thread_is_running();
+  if (r == 0)
     {
-      log_msg(LOG_ERR, "Failed to get peer UID: %s", strerror(-r));
+      log_msg(LOG_ERR, "chfn method called while already running!");
+      return -EPERM;
+    }
+
+  errno = 0; /* to find out if getpwnam succeed and there is no entry or if there was an error */
+  pw = getpwnam(p.name);
+  if (pw == NULL)
+    {
+      r = error_user_not_found(link, -1, p.name);
+      parameters_free(&p);
       return r;
     }
 
+  r = sd_varlink_get_peer_uid(link, &peer_uid);
+  if (r < 0)
+    return return_errno_error(link, "Get peer UID", r);
+
   r = sd_varlink_dispatch(p.link, parameters, dispatch_table, &p);
   if (r < 0)
     {
@@ -432,6 +448,24 @@ vl_method_chfn(sd_varlink *link, sd_json_variant *parameters,
 				SD_JSON_BUILD_PAIR_STRING("ErrorMsg", "No user name specified"));
     }
 
+  /* Don't change GECOS if query does not come from root
+     and result is not the one of the calling user */
+  if (!check_caller_perms(peer_uid, pw->pw_uid, NULL))
+    {
+      if (asprintf(&error, "Peer UID %u is not allowed to access data of '%s'",
+		   peer_uid, p.name) < 0)
+	error = NULL;
+      log_msg(LOG_ERR, "chfn: %s", stroom(error));
+      parameters_free(&p);
+      return sd_varlink_errorbo(link, "org.openSUSE.pwupd.InternalError",
+				SD_JSON_BUILD_PAIR_BOOLEAN("Success", false),
+				SD_JSON_BUILD_PAIR_STRING("ErrorMsg", stroom(error)));
+    }
+
+  /* Now that we know the caller is allowed to make changes,
+     check if the caller is allowed to change the single fields
+     and the new values are valid */
+
   if (p.full_name)
     {
       if (!may_change_field(peer_uid, 'f', &error))
@@ -518,15 +552,6 @@ vl_method_chfn(sd_varlink *link, sd_json_variant *parameters,
 	}
     }
 
-  errno = 0; /* to find out if getpwnam succeed and there is no entry or if there was an error */
-  pw = getpwnam(p.name);
-  if (pw == NULL)
-    {
-      r = error_user_not_found(link, -1, p.name);
-      parameters_free(&p);
-      return r;
-    }
-
   p.old_gecos = strdup(strempty(pw->pw_gecos));
   if (p.old_gecos == NULL)
     {
@@ -534,20 +559,6 @@ vl_method_chfn(sd_varlink *link, sd_json_variant *parameters,
       return return_errno_error(link, "strdup", -ENOMEM);
     }
 
-  /* Don't change GECOS if query does not come from root
-     and result is not the one of the calling user */
-  if (peer_uid != 0 && pw->pw_uid != peer_uid)
-    {
-      if (asprintf(&error, "Peer UID (%i) not 0 and peer UID not equal to UID",
-		   peer_uid) < 0)
-	error = NULL;
-      log_msg(LOG_ERR, "chfn: %s", stroom(error));
-      parameters_free(&p);
-      return sd_varlink_errorbo(link, "org.openSUSE.pwupd.InternalError",
-				SD_JSON_BUILD_PAIR_BOOLEAN("Success", false),
-				SD_JSON_BUILD_PAIR_STRING("ErrorMsg", stroom(error)));
-    }
-
   /* Run under the UID of the caller, else pam_unix will not ask
      for old password and pam_rootok will wrongly match. */
   if (peer_uid != 0)
@@ -563,16 +574,22 @@ vl_method_chfn(sd_varlink *link, sd_json_variant *parameters,
   r = pthread_create(&pam_thread, NULL, &run_pam_auth, &p);
   if (r != 0)
     return return_errno_error(link, "pthread_create", r);
+  pam_thread_is_valid = true;
 
   pthread_mutex_lock(&mut);
-  log_msg(LOG_DEBUG, "chfn: calling pthread_cond_wait");
-  pthread_cond_wait(&cond, &mut);
-  pthread_mutex_unlock(&mut);
-  log_msg(LOG_DEBUG, "chfn: pthread_cond_wait succeeded");
+  log_msg(LOG_DEBUG, "chfn: waiting for PAM thread");
+  while(!broadcast_called)
+    pthread_cond_wait(&cond, &mut);
+  broadcast_called = false;
 
   /* we need input from the user, quit method and send prompt back */
   if (send_v != NULL)
-    return sd_varlink_reply(link, send_v);
+    {
+      r = sd_varlink_reply(link, send_v);
+      pthread_mutex_unlock(&mut);
+      return r;
+    }
+  pthread_mutex_unlock(&mut);
 
   intptr_t *thread_res = NULL;
   r = pthread_join(pam_thread, (void **)&thread_res);
@@ -640,71 +657,62 @@ is_known_shell(const char *shell)
 }
 
 /* If the shell is completely invalid, print an error and
-   return 1. If root changes the shell, print only a warning.
+   return false. If root changes the shell, print only a warning.
    Only exception: Invalid characters are always not allowed.  */
-static int
-check_shell(const char *shell, uid_t uid, char **msg)
+static bool
+valid_shell(const char *shell, uid_t uid, char **msg)
 {
-  if (*shell != '/')
-    {
-      if (msg)
-	*msg = strdup("Shell must be a full path name.");
-      if (uid)
-        return 1;
-    }
-  else if (access(shell, F_OK) < 0)
-    {
-      if (msg)
-	{
-	  if (asprintf(msg, "'%s' does not exist.", shell) < 0)
-	    *msg = NULL;
-	}
-      if (uid)
-        return 1;
-    }
-  else if (access(shell, X_OK) < 0)
-    {
-      if (msg)
-	{
-	  if (asprintf(msg, "'%s' is not executable.", shell) < 0)
-	    *msg = NULL;
-	}
-      if (uid)
-        return 1;
-    }
+  assert(msg);
 
-  /* keep /etc/passwd clean. */
+  /* Keep /etc/passwd clean. This is always required, even for root. */
   for (size_t i = 0; i < strlen(shell); i++)
     {
       char c = shell[i];
-      if (c == ',' || c == ':' || c == '=' || c == '"' || c == '\n')
-        {
-	  if (msg)
-	    {
-	      if (asprintf(msg, "'%c' is not allowed.", c) < 0)
-		*msg = NULL;
-	    }
-          return 1;
-        }
+
       if (iscntrl (c))
         {
-          if (msg)
-	    *msg = strdup("Control characters are not allowed.");
-          return 1;
+	  *msg = strdup("Error: control characters are not allowed.");
+          return false;
+        }
+      if (c == ',' || c == ':' || c == '=' || c == '"')
+        {
+	  if (asprintf(msg, "Error: '%c' is not allowed.", c) < 0)
+	    *msg = NULL;
+          return false;
         }
     }
 
-  if (!is_known_shell(shell))
+  /* Check if the shell exists and is known.
+     Error for normal users, warning only for root */
+  if (*shell != '/')
     {
-      if (msg)
-	{
-	  if (asprintf(msg, "%s: '%s' is not listed as valid login shell.", uid?"Error":"Warning", shell) < 0)
-	    *msg = NULL;
-	}
+      if (asprintf(msg, "%s: shell must be a full path name.", uid?"Error":"Warning") < 0)
+	*msg = NULL;
       if (uid)
-	return 1;
+        return false;
     }
-  return 0;
+  else if (access(shell, F_OK) < 0)
+    {
+      if (asprintf(msg, "%s: '%s' does not exist.", uid?"Error":"Warning", shell) < 0)
+	*msg = NULL;
+      if (uid)
+        return false;
+    }
+  else if (access(shell, X_OK) < 0)
+    {
+      if (asprintf(msg, "%s: '%s' is not executable.", uid?"Error":"Warning", shell) < 0)
+	*msg = NULL;
+      if (uid)
+        return false;
+    }
+  else if (!is_known_shell(shell))
+    {
+      if (asprintf(msg, "%s: '%s' is not listed as valid login shell.", uid?"Error":"Warning", shell) < 0)
+	*msg = NULL;
+      if (uid)
+	return false;
+    }
+  return true;
 }
 
 static int
@@ -728,12 +736,10 @@ vl_method_chsh(sd_varlink *link, sd_json_variant *parameters,
     .content_passwd = NULL,
     .content_shadow = NULL,
     .link = link,
-    .run_as_user = 0,
   };
   static const sd_json_dispatch_field dispatch_table[] = {
     { "userName", SD_JSON_VARIANT_STRING,  sd_json_dispatch_string, offsetof(struct parameters, name),  SD_JSON_MANDATORY},
     { "shell",    SD_JSON_VARIANT_STRING,  sd_json_dispatch_string, offsetof(struct parameters, shell), SD_JSON_MANDATORY},
-    { "flags",    SD_JSON_VARIANT_INTEGER, sd_json_dispatch_int,    offsetof(struct parameters, flags), 0},
     {}
   };
   uid_t peer_uid;
@@ -741,19 +747,22 @@ vl_method_chsh(sd_varlink *link, sd_json_variant *parameters,
 
   log_msg(LOG_INFO, "Varlink method \"chsh\" called...");
 
+  /* If there is already a thread running, quit with error.
+     The conv method needs to be called to continue. */
+  r = thread_is_running();
+  if (r == 0)
+    {
+      log_msg(LOG_ERR, "chsh method called while already running!");
+      return -EPERM;
+    }
+
   r = sd_varlink_get_peer_uid(link, &peer_uid);
   if (r < 0)
-    {
-      log_msg(LOG_ERR, "Failed to get peer UID: %s", strerror(-r));
-      return r;
-    }
+    return return_errno_error(link, "Get peer UID", r);
 
   r = sd_varlink_dispatch(p.link, parameters, dispatch_table, &p);
   if (r < 0)
-    {
-      log_msg(LOG_ERR, "chsh request: varlink dispatch failed: %s", strerror(-r));
-      return r;
-    }
+    return return_errno_error(link, "chsh request: varlink dispatch", r);
 
   if (isempty(p.name))
     {
@@ -776,12 +785,12 @@ vl_method_chsh(sd_varlink *link, sd_json_variant *parameters,
 
   /* Don't change shell if query does not come from root
      and result is not the one of the calling user */
-  if (peer_uid != 0 && pw->pw_uid != peer_uid)
+  if (!check_caller_perms(peer_uid, pw->pw_uid, NULL))
     {
       _cleanup_free_ char *error = NULL;
 
-      if (asprintf(&error, "Peer UID (%i) not 0 and peer UID not equal to UID",
-		   peer_uid) < 0)
+      if (asprintf(&error, "Peer UID %u is not allowed to access data of '%s'",
+		   peer_uid, p.name) < 0)
 	error = NULL;
       log_msg(LOG_ERR, "chsh: %s", stroom(error));
       parameters_free(&p);
@@ -791,10 +800,9 @@ vl_method_chsh(sd_varlink *link, sd_json_variant *parameters,
     }
 
   _cleanup_free_ char *msg = NULL;
-  r = check_shell(p.shell, peer_uid, &msg);
-  if (r != 0)
+  if (!valid_shell(p.shell, peer_uid, &msg))
     {
-      log_msg(LOG_ERR, "chsh (check_shell): %s", stroom(msg));
+      log_msg(LOG_ERR, "%s", stroom(msg));
       parameters_free(&p);
       return sd_varlink_errorbo(link, "org.openSUSE.pwupd.InvalidShell",
 				SD_JSON_BUILD_PAIR_BOOLEAN("Success", false),
@@ -820,16 +828,22 @@ vl_method_chsh(sd_varlink *link, sd_json_variant *parameters,
   r = pthread_create(&pam_thread, NULL, &run_pam_auth, &p);
   if (r != 0)
     return return_errno_error(link, "pthread_create", r);
+  pam_thread_is_valid = true;
 
   pthread_mutex_lock(&mut);
-  log_msg(LOG_DEBUG, "chsh: calling pthread_cond_wait");
-  pthread_cond_wait(&cond, &mut);
-  pthread_mutex_unlock(&mut);
-  log_msg(LOG_DEBUG, "chsh: pthread_cond_wait succeeded");
+  log_msg(LOG_DEBUG, "chsh: waiting for PAM thread");
+  while(!broadcast_called)
+    pthread_cond_wait(&cond, &mut);
+  broadcast_called = false;
 
   /* we need input from the user, quit method and send prompt back */
   if (send_v != NULL)
-    return sd_varlink_reply(link, send_v);
+    {
+      r = sd_varlink_reply(link, send_v);
+      pthread_mutex_unlock(&mut);
+      return r;
+    }
+  pthread_mutex_unlock(&mut);
 
   intptr_t *thread_res = NULL;
   r = pthread_join(pam_thread, (void **)&thread_res);
@@ -879,7 +893,6 @@ run_pam_chauthtok(void *arg)
     .content_passwd = NULL,
     .content_shadow = NULL,
     .link = param->link,
-    .run_as_user = param->run_as_user,
   };
   const struct pam_conv conv = {
     varlink_conv,
@@ -896,21 +909,11 @@ run_pam_chauthtok(void *arg)
       return broadcast_and_return(r);
     }
 
-  if (p.run_as_user)
-    {
-      r = pam_putenv(pamh, "PAM_NO_ROOT=1");
-      if (r != PAM_SUCCESS)
-	{
-	  log_msg(LOG_ERR, "pam_putenv(\"PAM_NO_ROOT=1\") failed: %s", pam_strerror(NULL, r));
-	  return broadcast_and_return(r);
-	}
-    }
-
-  log_msg(LOG_DEBUG, "pam_chauthok(pamh, %i)", p.flags);
+  log_msg(LOG_DEBUG, "pam_chauthtok(pamh, %i)", p.flags);
   r = pam_chauthtok(pamh, p.flags);
   if (r != PAM_SUCCESS)
     {
-      pam_end (pamh, r);
+      pam_end(pamh, r);
       log_msg(LOG_ERR, "pam_chauthtok() failed: %s", pam_strerror(NULL, r));
       return broadcast_and_return(r);
     }
@@ -945,7 +948,6 @@ vl_method_chauthtok(sd_varlink *link, sd_json_variant *parameters,
     .content_passwd = NULL,
     .content_shadow = NULL,
     .link = link,
-    .run_as_user = 0,
   };
   static const sd_json_dispatch_field dispatch_table[] = {
     { "userName", SD_JSON_VARIANT_STRING,  sd_json_dispatch_string, offsetof(struct parameters, name),  SD_JSON_MANDATORY},
@@ -957,19 +959,22 @@ vl_method_chauthtok(sd_varlink *link, sd_json_variant *parameters,
 
   log_msg(LOG_INFO, "Varlink method \"chauthtok\" called...");
 
+  /* If there is already a thread running, quit with error.
+     The conv method needs to be called to continue. */
+  r = thread_is_running();
+  if (r == 0)
+    {
+      log_msg(LOG_ERR, "chauthtok method called while already running!");
+      return -EPERM;
+    }
+
   r = sd_varlink_get_peer_uid(link, &peer_uid);
   if (r < 0)
-    {
-      log_msg(LOG_ERR, "Failed to get peer UID: %s", strerror(-r));
-      return r;
-    }
+    return return_errno_error(link, "Get peer UID", r);
 
   r = sd_varlink_dispatch(p.link, parameters, dispatch_table, &p);
   if (r < 0)
-    {
-      log_msg(LOG_ERR, "chauthtok: varlink dispatch failed: %s", strerror(-r));
-      return r;
-    }
+    return return_errno_error(link, "chauthtok: varlink dispatch", r);
 
   if (isempty(p.name))
     {
@@ -992,12 +997,12 @@ vl_method_chauthtok(sd_varlink *link, sd_json_variant *parameters,
 
   /* Don't change password if query does not come from root
      and result is not the one of the calling user */
-  if (peer_uid != 0 && pw->pw_uid != peer_uid)
+  if (!check_caller_perms(peer_uid, pw->pw_uid, NULL))
     {
       _cleanup_free_ char *error = NULL;
 
-      if (asprintf(&error, "Peer UID (%i) not 0 and peer UID not equal to UID",
-		   peer_uid) < 0)
+      if (asprintf(&error, "Peer UID %u is not allowed to access data of '%s'",
+		   peer_uid, p.name) < 0)
 	error = NULL;
       log_msg(LOG_ERR, "chauthtok: %s", stroom(error));
       parameters_free(&p);
@@ -1021,21 +1026,27 @@ vl_method_chauthtok(sd_varlink *link, sd_json_variant *parameters,
   r = pthread_create(&pam_thread, NULL, &run_pam_chauthtok, &p);
   if (r != 0)
     return return_errno_error(link, "pthread_create", errno);
+  pam_thread_is_valid = true;
 
   pthread_mutex_lock(&mut);
-  log_msg(LOG_DEBUG, "chauthtok: calling pthread_cond_wait");
-  pthread_cond_wait(&cond, &mut);
-  pthread_mutex_unlock(&mut);
-  log_msg(LOG_DEBUG, "chauthtok: pthread_cond_wait succeeded");
+  log_msg(LOG_DEBUG, "chauthtok: waiting for PAM thread");
+  while(!broadcast_called)
+    pthread_cond_wait(&cond, &mut);
+  broadcast_called = false;
 
   /* we need input from the user, quit method and send prompt back */
   if (send_v != NULL)
-    return sd_varlink_reply(link, send_v);
+    {
+      r = sd_varlink_reply(link, send_v);
+      pthread_mutex_unlock(&mut);
+      return r;
+    }
+  pthread_mutex_unlock(&mut);
 
   intptr_t *thread_res = NULL;
   r = pthread_join(pam_thread, (void **)&thread_res);
   if (r != 0)
-    return return_errno_error(link, "pthread_joind", errno);
+    return return_errno_error(link, "pthread_join", errno);
 
   if (thread_res != PAM_SUCCESS)
     {
@@ -1071,7 +1082,6 @@ vl_method_conv(sd_varlink *link, sd_json_variant *parameters,
     .content_passwd = NULL,
     .content_shadow = NULL,
     .link = link,
-    .run_as_user = 0,
   };
   static const sd_json_dispatch_field dispatch_table[] = {
     { "response", SD_JSON_VARIANT_STRING, sd_json_dispatch_string,  offsetof(struct parameters, response),  SD_JSON_NULLABLE},
@@ -1082,43 +1092,42 @@ vl_method_conv(sd_varlink *link, sd_json_variant *parameters,
   log_msg(LOG_INFO, "Varlink method \"conv\" called...");
 
   /* make sure there is a pam_start() thread running! */
-  if (pam_thread != 0)
-    r = pthread_kill(pam_thread, 0);
-  else
-    r = ENOENT;
+  r = thread_is_running();
   if (r != 0)
     return return_errno_error(link, "Finding PAM thread", r);
 
   r = sd_varlink_dispatch(p.link, parameters, dispatch_table, &p);
   if (r < 0)
-    {
-      log_msg(LOG_ERR, "Conv request: varlink dispatch failed: %s", strerror(-r));
-      return r;
-    }
+    return return_errno_error(link, "Conv request: varlink dispatch", r);
 
   /* set pam_response */
   pthread_mutex_lock(&mut);
-  log_msg(LOG_DEBUG, "conv: set response and send cond_broadcast");
-  if (send_v != NULL)
-    sd_json_variant_unref(send_v);
-  send_v = NULL;
+  log_msg(LOG_DEBUG, "method_conv: set response and send cond_broadcast");
+  send_v = sd_json_variant_unref(send_v);
   if (p.response)
     answer = strdup(p.response);
   else
     answer = NULL;
+  got_answer = true;
+  /* inform PAM thread about answer */
   pthread_cond_broadcast(&cond);
   pthread_mutex_unlock(&mut);
 
   /* wait for next PAM_PROMPT_ECHO_* message or exit */
   pthread_mutex_lock(&mut);
-  log_msg(LOG_DEBUG, "conv: calling pthread_cond_wait");
-  pthread_cond_wait(&cond, &mut);
-  pthread_mutex_unlock(&mut);
-  log_msg(LOG_DEBUG, "conv: pthread_cond_wait succeeded");
+  log_msg(LOG_DEBUG, "method_conv: waiting for PAM thread");
+  while(!broadcast_called)
+    pthread_cond_wait(&cond, &mut);
+  broadcast_called = false;
 
   /* we need input from the user, quit method and send prompt back */
   if (send_v != NULL)
-    return sd_varlink_reply(link, send_v);
+    {
+      r = sd_varlink_reply(link, send_v);
+      pthread_mutex_unlock(&mut);
+      return r;
+    }
+  pthread_mutex_unlock(&mut);
 
   intptr_t *thread_res = NULL;
   r = pthread_join(pam_thread, (void **)&thread_res);
@@ -1142,9 +1151,9 @@ vl_method_conv(sd_varlink *link, sd_json_variant *parameters,
 }
 
 static int
-vl_method_update_pw_sp(sd_varlink *link, sd_json_variant *parameters,
-		       sd_varlink_method_flags_t _unused_(flags),
-		       void _unused_(*userdata))
+vl_method_UpdatePasswdShadow(sd_varlink *link, sd_json_variant *parameters,
+		             sd_varlink_method_flags_t _unused_(flags),
+		             void _unused_(*userdata))
 {
   _cleanup_(parameters_free) struct parameters p = {
     .name = NULL,
@@ -1160,7 +1169,6 @@ vl_method_update_pw_sp(sd_varlink *link, sd_json_variant *parameters,
     .content_passwd = NULL,
     .content_shadow = NULL,
     .link = link,
-    .run_as_user = 0,
   };
   static const sd_json_dispatch_field dispatch_table[] = {
     { "passwd",     SD_JSON_VARIANT_OBJECT,  sd_json_dispatch_variant, offsetof(struct parameters, content_passwd), SD_JSON_NULLABLE },
@@ -1235,8 +1243,10 @@ vl_method_update_pw_sp(sd_varlink *link, sd_json_variant *parameters,
 	return return_errno_error(link, "Parsing JSON shadow entry", r);
     }
 
-  /* XXX check that pw->pw_name and sp->sp_namp are identical if both
-     are provided */
+  /* Check that pw->pw_name and sp->sp_namp are identical if both are
+     provided */
+  if (pw && sp && !streq(pw->pw_name, sp->sp_namp))
+    return return_errno_error(link, "Check for identical account names", -EINVAL);
 
   if (pw)
     {
@@ -1301,7 +1311,7 @@ run_varlink (void)
 					  "org.openSUSE.pwupd.Chfn",               vl_method_chfn,
 					  "org.openSUSE.pwupd.Chsh",               vl_method_chsh,
 					  "org.openSUSE.pwupd.Conv",               vl_method_conv,
-					  "org.openSUSE.pwupd.UpdatePasswdShadow", vl_method_update_pw_sp,
+					  "org.openSUSE.pwupd.UpdatePasswdShadow", vl_method_UpdatePasswdShadow,
 					  "org.openSUSE.pwupd.GetEnvironment",     vl_method_get_environment,
 					  "org.openSUSE.pwupd.Ping",               vl_method_ping,
 					  "org.openSUSE.pwupd.Quit",               vl_method_quit,
